# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import warnings

warnings.warn("pymatgen.analysis.elasticity.tensors has been moved to "
              "pymatgen.core.tensors, please update dependencies accordingly. "
              "Links will removed in v2019.1.1.")
from pymatgen.core.tensors import *

